from plaid.api.api import API

link_token_field_names = [
    'user',
    'client_name',
    'products',
    'country_codes',
    'language',
    'redirect_uri',
    'android_package_name',
    'webhook',
    'link_customization_name',
    'access_token',
    'account_filters',
    'cross_app_item_add',
    'payment_initiation',
]


class LinkToken(API):
    '''Endpoints for managing link tokens.'''
    def create(self, configs):
        '''
        Create a Link token.

        :param dict configs: A required dictionary to configure the Link token.
        '''

        body = {}

        for field in link_token_field_names:
            body[field] = configs.get(field)

        return self.client.post('/link/token/create', body)



from plaid.api.api import API


class DepositSwitch(API):
    '''
    Deposit Switch endpoints.
    (`HTTP docs <coming soon>`__)
    '''

    def get(self, deposit_switch_id):
        '''
        Gets deposit switch information given a deposit switch id.

        :param  str deposit_switch_id:  ID of deposit switch to get.
        '''

        return self.client.post('/deposit_switch/get', {
            'deposit_switch_id': deposit_switch_id,
        })

    def create(self, target_account_id, target_access_token):
        '''
        Creates a deposit switch given target account id and target access
        token.
        (`HTTP docs <coming soon>`__)

        :param  str target_account_id:      The id of the bank account that
                                            the deposit switch will go to.
        :param  str target_access_token:    The access token to create the
                                            the deposit switch.
        '''
        return self.client.post('/deposit_switch/create', {
            'target_account_id': target_account_id,
            'target_access_token': target_access_token,
        })

    def create_token(self, deposit_switch_id):
        '''
        Creates a deposit switch token, which is used to initialize a deposit
        switch.
        (`HTTP docs <coming soon>`__)

        :param  str deposit_switch_id:  ID of deposit switch
        '''
        return self.client.post('/deposit_switch/token/create', {
            'deposit_switch_id': deposit_switch_id,
        })



from plaid.api.api import API


class Categories(API):
    '''
    Categories endpoints.
    (`HTTP docs <https://plaid.com/docs/api/#categories>`__)
    '''

    def get(self):
        '''Fetch all plaid categories.'''
        return self.client.post_public('/categories/get', {})



from plaid.api.api import API


class Income(API):
    '''
    Income endpoints.
    (`HTTP docs <https://plaid.com/docs/api/#income>`__)
    '''

    def get(self, access_token):
        '''
        Retrieve income data associated with an item.
        Adds the income product to the item if it does not already have it.

        :param str access_token:
        '''
        return self.client.post('/income/get', {
            'access_token': access_token,
        })



from plaid.api.api import API


class CreditDetails(API):
    '''
    Credit details endpoints.
    (`HTTP docs <https://plaid.com/docs/api/#credit-details>`__)
    '''

    def get(self, access_token):
        '''
        Retrieve credit details associated with an item.

        :param  str     access_token:
        '''
        return self.client.post('/credit_details/get', {
            'access_token': access_token,
        })



from tests.integration.util import (
    create_client,
)

access_token = None


def setup_module(module):
    client = create_client()
    # import an item
    at_response = client.Item.import_item(
        ['identity', 'auth'],
        {'user_id': 'user_good', 'auth_token': 'pass_good'},
        None)
    assert at_response['access_token'] is not None
    global access_token
    access_token = at_response['access_token']


def test_full_flow():
    client = create_client()
    # get account id
    account_id_response = client.Accounts.get(access_token)
    assert len(account_id_response['accounts']) != 0
    target_account_id = account_id_response['accounts'][0]['account_id']

    # create deposit switch
    create_response = client.DepositSwitch.create(
        target_account_id,
        access_token)
    assert create_response['deposit_switch_id'] is not None

    deposit_switch_id = create_response['deposit_switch_id']

    # get deposit switch
    get_response = client.DepositSwitch.get(deposit_switch_id)
    assert get_response['deposit_switch_id'] is not None
    assert get_response['target_item_id'] is not None
    assert get_response['target_account_id'] is not None
    assert get_response['date_created'] is not None
    assert get_response['state'] is not None

    # create deposit switch token
    token_response = client.DepositSwitch.create_token(deposit_switch_id)
    assert token_response['deposit_switch_token'] is not None
    assert token_response['deposit_switch_token_expiration_time'] is not None



from plaid.api.api import API


class AddToken(API):
    '''Endpoints for managing item add tokens.'''
    def create(self, user):
        '''
        Create a Link item add token.

        :param  dict user:  An optional dictionary with additional user data.
        '''
        return self.client.post('/item/add_token/create', {
            'user': user,
        })


class PublicToken(API):
    '''Endpoints for translating between public tokens and access tokens.'''

    def exchange(self, public_token):
        '''
        Exchange a Link public_token for an API access_token.
        (`HTTP docs <https://plaid.com/docs/api/#exchange-token-flow>`__)

        :param  str     public_token:
        '''
        return self.client.post('/item/public_token/exchange', {
            'public_token': public_token,
        })

    def create(self, access_token):
        '''
        Create a Link public_token for an API access_token.
        (`HTTP docs <https://plaid.com/docs/api/#create-public-token>`__)

        :param  str     access_token:
        '''
        return self.client.post('/item/public_token/create', {
            'access_token': access_token,
        })


class AccessToken(API):
    '''Access token endpoints.'''

    def invalidate(self, access_token):
        '''
        Rotate the access token for an item.
        (`HTTP docs <https://plaid.com/docs/api/#rotate-access-token>`__)

        :param  str     access_token:
        '''
        return self.client.post('/item/access_token/invalidate', {
            'access_token': access_token,
        })


class Webhook(API):
    '''Webhook endpoints.'''

    def update(self, access_token, webhook):
        '''
        Update the webhook for an Item.
        (`HTTP docs <https://plaid.com/docs/api/#update-webhook>`__)

        :param  str     access_token:
        :param  str     webhook: The URL of the webhook to associate.
        '''
        return self.client.post('./item/webhook/update', {
            'access_token': access_token,
            'webhook': webhook,
        })


class Item(API):
    '''
    Item endpoints.
    (`HTTP docs <https://plaid.com/docs/api/#item-management>`__)

    .. autoclass:: plaid.api.item.AddToken
        :members:
    .. autoclass:: plaid.api.item.AccessToken
        :members:
    .. autoclass:: plaid.api.item.PublicToken
        :members:
    .. autoclass:: plaid.api.item.Webhook
        :members:
    '''

    def __init__(self, client):
        super(Item, self).__init__(client)
        self.access_token = AccessToken(client)
        self.public_token = PublicToken(client)
        self.add_token = AddToken(client)
        self.webhook = Webhook(client)

    def get(self, access_token):
        '''
        Get information about the status of an item.
        (`HTTP docs <https://plaid.com/docs/api/#get-item>`__)

        :param  str     access_token:
        '''
        return self.client.post('/item/get', {
            'access_token': access_token,
        })

    def remove(self, access_token):
        '''
        Remove an item.
        (`HTTP docs <https://plaid.com/docs/api/#remove-an-item>`__)

        This also deactivates the access_token.

        :param  str     access_token:
        '''
        return self.client.post('/item/remove', {
            'access_token': access_token,
        })

    def import_item(self, products, user_auth, _options):
        '''
        Imports an item.
        (`HTTP docs coming soon`__)

        :param  [str]   initial_products:   List of products that the item
                                            will be enabled for.
        :param  dict    user_auth:          User authentication fields for the
                                            item.
        :param  dict    options:            Additional options.
        '''
        options = _options or {}

        return self.client.post('/item/import', {
            'products': products,
            'user_auth': user_auth,
            'options': options,
        })



from tests.integration.util import create_client


def test_all_payment_routes():
    client = create_client()

    # create recipient
    response = client.PaymentInitiation.create_recipient(
        'John Doe',
        'GB33BUKB20201555555555',
        {
            'street': ['street name 999'],
            'city': 'city',
            'postal_code': '99999',
            'country': 'GB',
        },
    )
    recipient_id = response['recipient_id']
    assert recipient_id is not None

    # get recipient
    response = client.PaymentInitiation.get_recipient(recipient_id)
    assert response['recipient_id'] is not None
    assert response['name'] is not None
    assert response['iban'] is not None
    assert response['address'] is not None

    # list recipients
    response = client.PaymentInitiation.list_recipients()
    assert response['recipients'] is not None
    assert len(response['recipients']) > 0

    # create payment
    response = client.PaymentInitiation.create_payment(
        recipient_id,
        'TestPayment',
        {
            'currency': 'GBP',
            'value': 100.00,
        },
    )
    payment_id = response['payment_id']
    assert payment_id is not None
    assert response['status'] is not None

    # create payment token
    response = client.PaymentInitiation.create_payment_token(
        payment_id,
    )
    assert response['payment_token'] is not None
    assert response['payment_token_expiration_time'] is not None

    # get payment
    response = client.PaymentInitiation.get_payment(payment_id)
    assert response['payment_id'] is not None
    assert response['payment_token'] is not None
    assert response['reference'] is not None
    assert response['amount'] is not None
    assert response['status'] is not None
    assert response['last_status_update'] is not None
    assert response['payment_token_expiration_time'] is not None
    assert response['recipient_id'] is not None

    # list payments
    response = client.PaymentInitiation.list_payments({'count': 10})
    assert response['payments'] is not None and len(response['payments']) > 0



from plaid.api.api import API


class Auth(API):
    '''Auth endpoints.'''

    def get(self,
            access_token,
            _options=None,
            account_ids=None):
        '''
        Retrieve account and routing numbers for checking and savings accounts.
        (`HTTP docs <https://plaid.com/docs/api/#auth>`__)

        :param  str     access_token:
        :param  [str]   account_ids:    A list of account_ids to retrieve for
                                        the item. Optional.
        '''
        options = _options or {}
        if account_ids is not None:
            options['account_ids'] = account_ids

        return self.client.post('/auth/get', {
            'access_token': access_token,
            'options': options,
        })



from plaid.errors import ItemError
from tests.integration.util import (
    create_client,
    SANDBOX_INSTITUTION
)

access_token = None


def setup_module(module):
    client = create_client()
    pt_response = client.Sandbox.public_token.create(
        SANDBOX_INSTITUTION, ['income'])
    exchange_response = client.Item.public_token.exchange(
        pt_response['public_token'])
    global access_token
    access_token = exchange_response['access_token']


def teardown_module(module):
    client = create_client()
    client.Item.remove(access_token)


def test_get():
    client = create_client()
    try:
        client.Income.get(access_token)
    except ItemError as ie:
        assert ie.code == u'PRODUCT_NOT_READY'



import time

from tests.integration.util import create_client


def test_link_token_create_required():
    client = create_client()

    # build the configs
    configs = {
        'user': {
            'client_user_id': str(time.time()),
        },
        'products': ["auth", "transactions"],
        'client_name': "Plaid Test",
        'country_codes': ['GB'],
        'language': 'en',
    }

    # create link token
    response = client.LinkToken.create(configs)

    # assert on response
    assert response['link_token'] is not None
    assert response['expiration'] is not None


def test_link_token_create_optional():
    client = create_client()

    # build the configs
    configs = {
        'user': {
            'client_user_id': str(time.time()),
        },
        'products': ["auth", "transactions"],
        'client_name': "Plaid Test",
        'country_codes': ['GB'],
        'language': 'en',
        'webhook': 'https://sample-webhook-uri.com',
        'link_customization_name': 'default',
        'account_filters': {
            'depository': {
                'account_subtypes': ['checking', 'savings'],
            },
        },
    }

    # create link token
    response = client.LinkToken.create(configs)

    # assert on response
    assert response['link_token'] is not None
    assert response['expiration'] is not None



from plaid.api.api import API


class Institutions(API):
    '''
    Institutions endpoints.
    (`HTTP docs <https://plaid.com/docs/api/#institutions>`__)
    '''

    def get(self, count, offset=0, _options=None):
        '''
        Fetch all Plaid institutions, using /institutions/all.

        :param  int     count:      Number of institutions to fetch.
        :param  int     offset:     Number of institutions to skip.
        '''

        options = _options or {}

        return self.client.post('/institutions/get', {
            'count': count,
            'offset': offset,
            'options': options,
        })

    def get_by_id(self, institution_id, _options=None):
        '''
        Fetch a single institution by id.

        :param  str     institution_id:
        '''
        options = _options or {}

        return self.client.post('/institutions/get_by_id', {
            'institution_id': institution_id,
            'options': options,
        })

    def search(self, query, _options=None, products=None):
        '''
        Search all institutions by name.

        :param  str      query:     Query against the full list of
                                    institutions.
        :param  [str]    products:  Filter FIs by available products. Optional.
        '''
        options = _options or {}

        return self.client.post('/institutions/search', {
            'query': query,
            'products': products,
            'options': options,
        })



import time
from plaid.errors import AssetReportError
from tests.integration.util import (
    create_client,
    SANDBOX_INSTITUTION,
)


access_token = None


def setup_module(module):
    client = create_client()
    pt_response = client.Sandbox.public_token.create(
        SANDBOX_INSTITUTION, ['assets'])
    exchange_response = client.Item.public_token.exchange(
        pt_response['public_token'])
    global access_token
    access_token = exchange_response['access_token']


def teardown_module(module):
    client = create_client()
    client.Item.remove(access_token)


def test_full_flow():
    client = create_client()

    # create an asset report for one item
    options = {
        'client_report_id': '123',
        'webhook': 'https://www.example.com',
        'user': {
            'client_user_id': '789',
            'first_name': 'Jane',
            'middle_name': 'Leah',
            'last_name': 'Doe',
            'ssn': '123-45-6789',
            'phone_number': '(555) 123-4567',
            'email': 'jane.doe@example.com',
        }
    }
    response = client.AssetReport.create(
        [access_token],
        days_requested=60,
        options=options)
    asset_report_token = response['asset_report_token']
    asset_report_id = response['asset_report_id']
    assert asset_report_token is not None
    assert asset_report_id is not None

    # retrieve the asset report
    response = poll_for_asset_report(client, asset_report_token)
    report = response['report']
    assert report is not None

    # retrieve the asset report as an Asset Report with Insights
    response = client.AssetReport.get(asset_report_token, True)
    report = response['report']
    assert report is not None

    # The transactions in an Asset Report with Insights should have a non-null
    # `name` (when available).
    assert (
        name_exists_for_some_transaction(report))

    # retrieve the asset report as a PDF
    pdf = client.AssetReport.get_pdf(asset_report_token)
    assert pdf is not None

    # create a filtered copy of the asset report
    account_ids_to_exclude = [report['items'][0]['accounts'][0]['account_id']]
    response = client.AssetReport.filter(
        asset_report_token, account_ids_to_exclude)
    assert response['asset_report_token'] is not None

    # create a refreshed copy of the asset report
    response = client.AssetReport.refresh(asset_report_token, 10)
    assert response['asset_report_token'] is not None

    # create an audit copy
    response = client.AssetReport.audit_copy.create(
        asset_report_token,
        client.client_id)
    audit_copy_token = response['audit_copy_token']
    assert audit_copy_token is not None

    # get the audit copy
    response = client.AssetReport.audit_copy.get(audit_copy_token)
    audit_copy = response['report']
    assert audit_copy is not None

    # remove the audit copy token
    response = client.AssetReport.audit_copy.remove(audit_copy_token)
    removed = response['removed']
    assert removed

    # remove the asset report
    response = client.AssetReport.remove(asset_report_token)
    removed = response['removed']
    assert removed


def name_exists_for_some_transaction(report):
    for account in report['items'][0]['accounts']:
        if len(account['transactions']) > 0:
            return (account['transactions'][0]['name'] is not None)

    return False


def poll_for_asset_report(client, asset_report_token, retries=20):
    try:
        return client.AssetReport.get(asset_report_token)
    except AssetReportError as e:
        if e.code == 'PRODUCT_NOT_READY' and retries > 0:
            time.sleep(1)
            return poll_for_asset_report(
                client,
                asset_report_token,
                retries - 1)

        raise e



'''
Webhook tests
'''

from tests.integration.util import (
    create_client,
    WEBHOOK_VERIFICATION_KEY_ID
)


def test_get_webhook_verification_key():
    client = create_client()
    pt_response = client.Webhooks.get_verification_key(
        WEBHOOK_VERIFICATION_KEY_ID
    )
    assert pt_response['key'] is not None
    assert pt_response['key']['alg'] is not None
    assert pt_response['key']['created_at'] is not None
    assert pt_response['key']['crv'] is not None
    assert pt_response['key']['kid'] is not None
    assert pt_response['key']['kty'] is not None
    assert pt_response['key']['use'] is not None
    assert pt_response['key']['x'] is not None
    assert pt_response['key']['y'] is not None



from tests.integration.util import (
    create_client,
    SANDBOX_INSTITUTION,
    SANDBOX_INSTITUTION_NAME,
)


def test_get():
    client = create_client()
    response = client.Institutions.get(3, offset=1)
    assert len(response['institutions']) == 3


def test_get_with_include_optional_metadata():
    client = create_client()
    response = client.Institutions.get(3, offset=1, _options={
        'include_optional_metadata': True,
    })
    assert len(response['institutions']) == 3
    assert len(response['institutions'][0]['url']) > 0
    assert len(response['institutions'][0]['primary_color']) > 0

    assert len(response['institutions'][1]['url']) > 0
    assert len(response['institutions'][1]['primary_color']) > 0

    assert len(response['institutions'][2]['url']) > 0
    assert len(response['institutions'][2]['primary_color']) > 0


def test_get_by_id():
    client = create_client()
    response = client.Institutions.get_by_id(SANDBOX_INSTITUTION)
    assert response['institution']['institution_id'] == SANDBOX_INSTITUTION


def test_get_by_id_with_include_optional_metadata():
    client = create_client()
    response = client.Institutions.get_by_id(SANDBOX_INSTITUTION, _options={
        'include_optional_metadata': True,
    })
    assert response['institution']['institution_id'] == SANDBOX_INSTITUTION


def test_search():
    client = create_client()
    response = client.Institutions.search(SANDBOX_INSTITUTION_NAME)
    assert len(response['institutions']) >= 1


def test_search_with_products():
    client = create_client()
    response = client.Institutions.search(
        SANDBOX_INSTITUTION_NAME, products=['transactions'])
    assert len(response['institutions']) >= 1


def test_search_with_include_optional_metadata():
    client = create_client()
    response = client.Institutions.search(
        SANDBOX_INSTITUTION_NAME, _options={
            'include_optional_metadata': True,
        })
    assert len(response['institutions']) >= 1



from plaid.api.api import API


class Webhooks(API):
    '''
    Webhooks endpoints.
    (`HTTP docs <https://plaid.com/docs/#webhook-verification>`__)
    '''

    def get_verification_key(self,
                             key_id):
        '''
        Retrieve webhook verification key.

        :param  str     key_id: ID of key from Plaid-Verification header
        '''
        return self.client.post('/webhook_verification_key/get', {
            'key_id': key_id,
        })



from plaid.errors import APIError, PlaidError


def test_from_response():
    response = {
        'display_message': None,
        'error_type': 'API_ERROR',
        'error_code': 'INTERNAL_SERVER_ERROR',
        'error_message': 'an unexpected error occurred',
        'request_id': 'abc123',
        'causes': [
            {
                'error_type': 'API_ERROR',
                'error_code': 'INTERNAL_SERVER_ERROR',
                'error_message': 'an unexpected error occurred',
                'item_id': '456',
            },
        ],
    }

    error = PlaidError.from_response(response)
    assert isinstance(error, APIError)
    assert error.code == 'INTERNAL_SERVER_ERROR'
    assert error.message == 'an unexpected error occurred'

    assert len(error.causes) == 1
    cause = error.causes[0]
    assert cause.type == 'API_ERROR'
    assert cause.code == 'INTERNAL_SERVER_ERROR'
    assert cause.message == 'an unexpected error occurred'
    assert cause.item_id == '456'



from plaid.api.api import API


class AssetReport(API):
    '''Assets endpoints.
    (`HTTP docs <https://plaid.com/docs/api/#assets>`__)'''

    def __init__(self, client):
        super(AssetReport, self).__init__(client)
        self.audit_copy = AuditCopy(client)

    def create(self,
               access_tokens,
               days_requested,
               options=None):
        '''
        Create an asset report.

        :param  [str]   access_tokens:  A list of access tokens, one token for
                                        each Item to be included in the Asset
                                        Report.
        :param  int     days_requested: Days of transaction history requested
                                        to be included in the Asset Report.
        :param  dict    options:        An optional dictionary. For more
                                        information on the options object, see
                                        the documentation site listed above.
        '''
        options = options or {}

        return self.client.post('/asset_report/create', {
            'access_tokens': access_tokens,
            'days_requested': days_requested,
            'options': options,
        })

    def filter(self,
               asset_report_token,
               account_ids_to_exclude):
        '''
        Create a new, filtered asset report based on an existing asset report.

        :param  str   asset_report_token:       The existing Asset Report's
                                                asset report token.
        :param  [str] account_ids_to_exclude:   A list of account IDs to
                                                exclude from the new Asset
                                                Report.
        '''
        return self.client.post('/asset_report/filter', {
            'asset_report_token': asset_report_token,
            'account_ids_to_exclude': account_ids_to_exclude,
        })

    def refresh(self,
                asset_report_token,
                days_requested,
                options=None):
        '''
        Create a new, refreshed asset report based on an existing asset report.

        :param  str   asset_report_token:   The existing Asset Report's asset
                                            report token.
        :param  int   days_requested:       Days of transaction history
                                            requested to be included in the
                                            Asset Report.
        :param  dict  options:              An optional dictionary. This is the
                                            same object used in `create`.
        '''
        options = options or {}

        return self.client.post('/asset_report/refresh', {
            'asset_report_token': asset_report_token,
            'days_requested': days_requested,
            'options': options,
        })

    def get(self,
            asset_report_token,
            include_insights=False):
        '''
        Retrieves an asset report.

        :param  str   asset_report_token:   The asset report token for the
                                            asset report you created.
        :param  bool  include_insights:     An optional boolean specifying
                                            whether we should retrieve the
                                            report as an Asset Report with
                                            Insights. For more, see
                                            https://plaid.com/docs/#retrieve-json-report-request.
        '''

        return self.client.post('/asset_report/get', {
            'asset_report_token': asset_report_token,
            'include_insights': include_insights,
        })

    def get_pdf(self,
                asset_report_token):
        '''
        Retrieves an asset report in the PDF format.

        :param  str   asset_report_token:   The asset report token for the
                                            asset report you created.
        '''

        return self.client.post('/asset_report/pdf/get', {
            'asset_report_token': asset_report_token,
        }, is_json=False)

    def remove(self,
               asset_report_token):
        '''
        Retrieves an asset report in the PDF format.

        :param  str   asset_report_token:   The asset report token for the
                                            asset report you want to remove.
        '''

        return self.client.post('/asset_report/remove', {
            'asset_report_token': asset_report_token,
        })


class AuditCopy(API):
    '''Audit copy endpoints. Use this class via the `audit_copy` member on the
    `AssetReport` class.'''

    def create(self,
               asset_report_token,
               auditor_id):
        '''
        Creates an audit copy.

        :param  str   asset_report_token:   The asset report token for the
                                            asset report you created.
        :param  str   auditor_id:           The ID of the third party with
                                            which you want to share the asset
                                            report.

        '''

        return self.client.post('/asset_report/audit_copy/create', {
            'asset_report_token': asset_report_token,
            'auditor_id': auditor_id,
        })

    def get(self,
            audit_copy_token):
        '''
        Retrieves an audit copy.

        :param  str   audit_copy_token:     The audit copy token for the audit
                                            copy you want to retrieve.
        '''

        return self.client.post('/asset_report/audit_copy/get', {
            'audit_copy_token': audit_copy_token,
        })

    def remove(self,
               audit_copy_token):
        '''
        Removes an audit copy.

        :param  str   audit_copy_token:     The audit copy token for the
                                            audit copy you want to remove.

        '''

        return self.client.post('/asset_report/audit_copy/remove', {
            'audit_copy_token': audit_copy_token,
        })



from plaid.api.api import API


class Processor(API):
    '''Endpoints for creating processor tokens.'''

    def tokenCreate(self, access_token, account_id, processor):
        '''
        Create a processor token for a given processor and account ID

        :param  str     access_token:
        :param  str     account_id:
        :param  str     processor:
        '''
        endpoint = 'processor/token/create'
        options = {
            'access_token': access_token,
            'account_id': account_id,
            'processor': processor
        }

        if processor == 'stripe':
            endpoint = '/processor/stripe/bank_account_token/create'
            del options['processor']
        elif processor == 'apex':
            endpoint = '/processor/apex/processor_token/create'
            del options['processor']

        return self.client.post(endpoint, options)

    def stripeBankAccountTokenCreate(self, access_token, account_id):
        '''
        Create a Stripe bank_account token for a given account ID
        (`HTTP docs <https://plaid.com/docs/link/stripe>`__)

        :param  str     access_token:
        :param  str     account_id:
        '''
        return self.client.post('/processor/stripe/bank_account_token/create',
                                {
                                    'access_token': access_token,
                                    'account_id': account_id,
                                })

    def dwollaBankAccountTokenCreate(self, access_token, account_id):
        '''
        Create a Dwolla processor token for a given account ID
        (`HTTP docs <https://plaid.com/docs/link/dwolla>`__)

        :param  str     access_token:
        :param  str     account_id:
        '''
        return self.client.post('/processor/dwolla/processor_token/create',
                                {
                                    'access_token': access_token,
                                    'account_id': account_id,
                                })



import json
from functools import partial

import requests

from plaid.errors import PlaidError
from plaid.version import __version__


ALLOWED_METHODS = {'post'}
DEFAULT_TIMEOUT = 600  # 10 minutes

try:
    from json.decoder import JSONDecodeError
except ImportError:
    # json parsing throws a ValueError in python2
    JSONDecodeError = ValueError


def _requests_http_request(
        url,
        method,
        data,
        headers,
        timeout=DEFAULT_TIMEOUT):
    normalized_method = method.lower()
    headers.update({'User-Agent': 'Plaid Python v{}'.format(__version__)})
    if normalized_method in ALLOWED_METHODS:
        return getattr(requests, normalized_method)(
            url,
            json=data,
            headers=headers,
            timeout=timeout,
        )
    else:
        raise Exception(
            'Invalid request method {}'.format(method)
        )


def _http_request(
        url,
        method=None,
        data=None,
        headers=None,
        timeout=DEFAULT_TIMEOUT,
        is_json=True):
    response = _requests_http_request(
        url,
        method,
        data or {},
        headers or {},
        timeout)

    if is_json or response.headers['Content-Type'] == 'application/json':
        try:
            response_body = json.loads(response.text)
        except JSONDecodeError:
            raise PlaidError.from_response({
                'error_message': response.text,
                'error_type': 'API_ERROR',
                'error_code': 'INTERNAL_SERVER_ERROR',
                'display_message': None,
                'request_id': '',
                'causes': [],
            })
        if response_body.get('error_type'):
            raise PlaidError.from_response(response_body)
        else:
            return response_body
    else:
        return response.content


# helpers to simplify partial function application
post_request = partial(_http_request, method='POST')



import time

from plaid.errors import ItemError
from tests.integration.util import (
    create_client,
    SANDBOX_INSTITUTION,
)

access_token = None


def setup_module(module):
    client = create_client()
    pt_response = client.Sandbox.public_token.create(
        SANDBOX_INSTITUTION, ['investments'],
        transactions__start_date='2018-01-01',
        transactions__end_date='2019-01-01',
    )
    exchange_response = client.Item.public_token.exchange(
        pt_response['public_token'])
    global access_token
    access_token = exchange_response['access_token']


def get_investment_transactions_with_retries(client,
                                             access_token,
                                             start_date,
                                             end_date,
                                             account_ids=None,
                                             count=None,
                                             offset=None,
                                             num_retries=5):
    response = None
    for i in range(num_retries):
        try:
            response = \
                client.InvestmentTransactions.get(access_token,
                                                  start_date,
                                                  end_date,
                                                  account_ids=account_ids,
                                                  count=count,
                                                  offset=offset)
        except ItemError as ie:
            if ie.code == u'PRODUCT_NOT_READY':
                time.sleep(5)
                continue
            else:
                raise ie
        break
    return response


def teardown_module(module):
    client = create_client()
    client.Item.remove(access_token)


def test_get():
    client = create_client()

    response = get_investment_transactions_with_retries(client,
                                                        access_token,
                                                        '2018-01-01',
                                                        '2019-01-01',
                                                        num_retries=5)
    assert response['item'] is not None
    assert response['accounts'] is not None
    assert response['securities'] is not None
    assert response['investment_transactions'] is not None
    assert response['total_investment_transactions'] is not None

    # get transactions for selected accounts
    account_id = response['accounts'][0]['account_id']
    response = \
        get_investment_transactions_with_retries(client,
                                                 access_token,
                                                 '2018-06-01',
                                                 '2019-06-01',
                                                 account_ids=[account_id],
                                                 num_retries=5)
    assert response['investment_transactions'] is not None


def test_get_with_options():
    client = create_client()
    response = get_investment_transactions_with_retries(client,
                                                        access_token,
                                                        '2018-06-01',
                                                        '2019-06-01',
                                                        count=2,
                                                        offset=1)
    assert len(response['investment_transactions']) == 2



from plaid.api.api import API


class Balance(API):
    '''Accounts balance endpoint.'''

    def get(self,
            access_token,
            _options=None,
            account_ids=None):
        '''
        Retrieve real-time balance information for accounts.

        :param  str     access_token:
        :param  [str]   account_ids:    A list of account_ids to retrieve for
                                        the item. Optional.
        '''
        options = _options or {}
        if account_ids is not None:
            options['account_ids'] = account_ids

        return self.client.post('/accounts/balance/get', {
            'access_token': access_token,
            'options': options,
        })


class Accounts(API):
    '''
    Accounts endpoints.
    (`HTTP docs <https://plaid.com/docs/api/#accounts>`__)

    .. autoclass:: plaid.api.accounts.Balance
        :members:
    '''

    def __init__(self, client):
        super(Accounts, self).__init__(client)
        self.balance = Balance(client)

    def get(self,
            access_token,
            _options=None,
            account_ids=None):
        '''
        Retrieve high-level account information for an Item.

        :param  str     access_token:
        :param  [str]   account_ids:    A list of account_ids to retrieve for
                                        the item. Optional.
        '''
        options = _options or {}
        if account_ids is not None:
            options['account_ids'] = account_ids

        return self.client.post('/accounts/get', {
            'access_token': access_token,
            'options': options,
        })



'''
Client.Item.* tests.

Note that when using many of these methods in production, it would be necessary
to handle error cases (e.g. ``INVALID_CREDENTIAL``). However, we don't do that
here since any errors will automatically be marked as failures by the test
runner.
'''

import time
from contextlib import contextmanager

from tests.integration.util import (
    create_client,
    SANDBOX_INSTITUTION,
)


# Ensure that any items created are also removed
@contextmanager
def ensure_item_removed(access_token):
    try:
        yield
    finally:
        create_client().Item.remove(access_token)


def test_get():
    client = create_client()
    pt_response = client.Sandbox.public_token.create(
        SANDBOX_INSTITUTION, ['transactions'])
    exchange_response = client.Item.public_token.exchange(
        pt_response['public_token'])

    with ensure_item_removed(exchange_response['access_token']):
        get_response = client.Item.get(exchange_response['access_token'])
        assert get_response['item'] is not None


def test_remove():
    client = create_client()
    pt_response = client.Sandbox.public_token.create(
        SANDBOX_INSTITUTION, ['transactions'])
    exchange_response = client.Item.public_token.exchange(
        pt_response['public_token'])

    remove_response = client.Item.remove(exchange_response['access_token'])
    assert remove_response['removed']


def test_import():
    client = create_client()
    at_response = client.Item.import_item(
        ['identity', 'auth'],
        {'user_id': 'user_good', 'auth_token': 'pass_good'},
        None)
    assert at_response['access_token'] is not None


def test_add_token():
    client = create_client()
    create_response = client.Item.add_token.create({
        'client_user_id': str(time.time()),
    })
    assert create_response['add_token'] is not None
    assert create_response['expiration'] is not None


def test_add_token_with_user():
    client = create_client()
    user = {
        'client_user_id': str(time.time()),
        'phone_number': '+1 415 555 0123',
        'email_address': 'example@plaid.com',
        'phone_number_verified_time': '2020-01-01T00:00:00Z',
    }
    create_response = client.Item.add_token.create(user=user)
    assert create_response['add_token'] is not None
    assert create_response['expiration'] is not None


def test_public_token():
    client = create_client()
    pt_response = client.Sandbox.public_token.create(
        SANDBOX_INSTITUTION, ['transactions'])
    exchange_response = client.Item.public_token.exchange(
        pt_response['public_token'])
    with ensure_item_removed(exchange_response['access_token']):
        assert pt_response['public_token'] is not None
        assert exchange_response['access_token'] is not None


def test_sandbox_public_token():
    client = create_client()
    create_response = client.Sandbox.public_token.create(
        SANDBOX_INSTITUTION, ['transactions'])
    assert create_response['public_token'] is not None

    # public token -> access token
    exchange_response = client.Item.public_token.exchange(
        create_response['public_token'])
    assert exchange_response['access_token'] is not None


def test_sandbox_fire_webhook():
    client = create_client()
    create_response = client.Sandbox.public_token.create(
        SANDBOX_INSTITUTION, ['transactions'],
        webhook='https://plaid.com/foo/bar/hook')
    assert create_response['public_token'] is not None

    # public token -> access token
    exchange_response = client.Item.public_token.exchange(
        create_response['public_token'])
    assert exchange_response['access_token'] is not None

    # fire webhook
    fire_webhook_response = client.Sandbox.item.fire_webhook(
        exchange_response['access_token'],
        'DEFAULT_UPDATE'
    )
    assert fire_webhook_response['webhook_fired'] is True


def test_access_token_invalidate():
    client = create_client()
    pt_response = client.Sandbox.public_token.create(
        SANDBOX_INSTITUTION, ['transactions'])
    exchange_response = client.Item.public_token.exchange(
        pt_response['public_token'])

    try:
        invalidate_response = client.Item.access_token.invalidate(
            exchange_response['access_token'])
        with ensure_item_removed(invalidate_response['new_access_token']):
            assert invalidate_response['new_access_token'] is not None
    except Exception:
        with ensure_item_removed(exchange_response['access_token']):
            raise


def test_webhook_update():
    client = create_client()
    pt_response = client.Sandbox.public_token.create(
        SANDBOX_INSTITUTION, ['transactions'])
    exchange_response = client.Item.public_token.exchange(
        pt_response['public_token'])

    with ensure_item_removed(exchange_response['access_token']):
        webhook_response = client.Item.webhook.update(
            exchange_response['access_token'],
            'https://plaid.com/webhook-test')
        assert (webhook_response['item']['webhook'] ==
                'https://plaid.com/webhook-test')



from plaid.api.api import API


class Identity(API):
    '''
    Identity endpoints.
    (`HTTP docs <https://plaid.com/docs/api/#identity>`__)
    '''

    def get(self,
            access_token):
        '''
        Retrieve information about an item.

        :param  str     access_token:
        '''
        return self.client.post('/identity/get', {
            'access_token': access_token,
        })



'''
Client.Processor.* tests.
'''
import pytest

from plaid.errors import InvalidRequestError

from tests.integration.util import (
    create_client
)


def test_create_token():
    client = create_client()
    # Just test the failure case - behavior here depends on the API keys used
    with pytest.raises(InvalidRequestError) as e:
        client.Processor.tokenCreate(
            'fakeAccessToken', 'fakeAccountId', 'fakeProcessor')
        assert e.code == 'INVALID_INPUT'


def test_stripe_processor_token():
    client = create_client()
    # Just test the failure case - behavior here depends on the API keys used
    with pytest.raises(InvalidRequestError) as e:
        client.Processor.stripeBankAccountTokenCreate(
            'fakeAccessToken', 'fakeAccountId')
        assert e.code == 'INVALID_INPUT'


def test_dwolla_processor_token():
    client = create_client()
    # Just test the failure case - behavior here depends on the API keys used
    with pytest.raises(InvalidRequestError) as e:
        client.Processor.dwollaBankAccountTokenCreate(
            'fakeAccessToken', 'fakeAccountId')
        assert e.code == 'INVALID_INPUT'



import warnings

from plaid.api import (
    Accounts,
    AssetReport,
    Auth,
    Categories,
    CreditDetails,
    DepositSwitch,
    Holdings,
    Identity,
    Income,
    Institutions,
    InvestmentTransactions,
    Item,
    Liabilities,
    LinkToken,
    PaymentInitiation,
    Processor,
    Sandbox,
    Transactions,
    Webhooks,
)

from plaid.internal.requester import DEFAULT_TIMEOUT, post_request
from plaid.internal.utils import urljoin


class Client(object):
    '''
    Python Plaid API client.

    See official documentation at: https://plaid.com/docs.

    All of the endpoints documented under the ``plaid.api``
    module may be called from a ``plaid.Client`` instance.
    '''

    def __init__(self,
                 client_id=None,
                 secret=None,
                 environment=None,
                 suppress_warnings=False,
                 timeout=DEFAULT_TIMEOUT,
                 api_version=None,
                 client_app=None):
        '''
        Initialize a client with credentials.

        :param  str     client_id:          Your Plaid client ID
        :arg    str     secret:             Your Plaid secret
        :arg    str     environment:        One of ``sandbox``,
                                            ``development``, or ``production``.
        :arg    bool    suppress_warnings:  Suppress Plaid warnings.
        :arg    int     timeout:            Timeout for API requests.
        :arg    str     api_version:        API version to use for requests
        :arg    str     client_app:         Internal header to include
                                                in requests
        '''
        self.client_id = client_id
        self.secret = secret
        self.environment = environment
        self.suppress_warnings = suppress_warnings
        self.timeout = timeout
        self.api_version = api_version
        self.client_app = client_app

        if self.environment == 'development' and not self.suppress_warnings:
            warnings.warn('''
                Development is not intended for production usage.
                Swap out url for https://production.plaid.com
                via Client.config before switching to production
            ''')

        # Mirror the HTTP API hierarchy
        self.Accounts = Accounts(self)
        self.AssetReport = AssetReport(self)
        self.Auth = Auth(self)
        self.Categories = Categories(self)
        self.CreditDetails = CreditDetails(self)
        self.DepositSwitch = DepositSwitch(self)
        self.Holdings = Holdings(self)
        self.Identity = Identity(self)
        self.Income = Income(self)
        self.Institutions = Institutions(self)
        self.InvestmentTransactions = InvestmentTransactions(self)
        self.Item = Item(self)
        self.Liabilities = Liabilities(self)
        self.LinkToken = LinkToken(self)
        self.PaymentInitiation = PaymentInitiation(self)
        self.Processor = Processor(self)
        self.Sandbox = Sandbox(self)
        self.Transactions = Transactions(self)
        self.Webhooks = Webhooks(self)

    def post(self, path, data, is_json=True):
        '''Make a post request with client_id and secret key.'''
        post_data = {
            'client_id': self.client_id,
            'secret': self.secret,
        }
        post_data.update(data)
        return self._post(path, post_data, is_json)

    def post_public(self, path, data, is_json=True):
        '''Make a post request requiring no auth.'''
        return self._post(path, data, is_json)

    def _post(self, path, data, is_json):
        headers = {}
        if self.api_version is not None:
            headers['Plaid-Version'] = self.api_version
        if self.client_app is not None:
            headers['Plaid-Client-App'] = self.client_app
        return post_request(
            urljoin('https://' + self.environment + '.plaid.com', path),
            data=data,
            timeout=self.timeout,
            is_json=is_json,
            headers=headers,
        )



'''Shared objects for integration testing.'''

import os

from plaid import Client


def create_client():
    '''Create a new client for testing.'''
    return Client(
        client_id=os.environ['CLIENT_ID'],
        secret=os.environ['SECRET'],
        environment='sandbox',
        api_version="2019-05-29",
        client_app="plaid-python-unit-tests"
    )


SANDBOX_INSTITUTION = 'ins_109508'
SANDBOX_INSTITUTION_NAME = 'First Platypus Bank'

SANDBOX_INSTITUTIONS = [
    'ins_109508',
    'ins_109509',
    'ins_109510',
    'ins_109511',
    'ins_109512',
]

WEBHOOK_VERIFICATION_KEY_ID = '6c5516e1-92dc-479e-a8ff-5a51992e0001'



from plaid.api.api import API


class Item(API):
    '''Sandbox item endpoints.'''

    def reset_login(self, access_token):
        '''
        Put an item into an ITEM_LOGIN_REQUIRED error state.

        :param  str     access_token:
        '''
        return self.client.post('/sandbox/item/reset_login', {
            'access_token': access_token,
        })

    def fire_webhook(self, access_token, webhook_code):
        '''
        Fire a webhook for an item

        :param  str     access_token:
        :param  str     webhook_code:
        '''
        return self.client.post('/sandbox/item/fire_webhook', {
            'access_token': access_token,
            'webhook_code': webhook_code,
        })

    def set_verification_status(self, access_token, account_id,
                                verification_status):
        '''
        Set verification status for an item created via the
        automated microdeposits flow

        :param  str     access_token:
        :param  str     account_id:
        :param  str     verification_status:
        '''
        return self.client.post('/sandbox/item/set_verification_status', {
            'access_token': access_token,
            'account_id': account_id,
            'verification_status': verification_status,
        })


class PublicToken(API):
    '''Sandbox public token endpoints.'''

    def create(self,
               institution_id,
               initial_products,
               _options=None,
               webhook=None,
               transactions__start_date=None,
               transactions__end_date=None,
               ):
        '''
        Generate a public token for sandbox testing.

        :param  str     institution_id:

        :param  [str]   initial_products:

        :param  str     webhook:
        '''
        options = _options or {}

        if webhook is not None:
            options['webhook'] = webhook

        transaction_options = {}
        transaction_options.update(options.get('transactions', {}))
        if transactions__start_date is not None:
            transaction_options['start_date'] = transactions__start_date
        if transactions__end_date is not None:
            transaction_options['end_date'] = transactions__end_date
        if transaction_options:
            options['transactions'] = transaction_options

        return self.client.post('/sandbox/public_token/create', {
            'institution_id': institution_id,
            'initial_products': initial_products,
            'options': options,
        })


class Sandbox(API):
    '''
    Sandbox-only endpoints.
    (`HTTP docs <https://plaid.com/docs/api/#sandbox>`__)

    These endpoints may not be used in other environments.

    .. autoclass:: plaid.api.sandbox.Item
        :members:
    '''

    def __init__(self, client):
        super(Sandbox, self).__init__(client)
        self.item = Item(client)
        self.public_token = PublicToken(client)



import re
from setuptools import setup, find_packages

with open('plaid/version.py', 'r') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        f.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Cannot find version information')

url = 'https://github.com/plaid/plaid-python'

setup(
    name='plaid-python',
    version=version,
    description='Python client library for the Plaid API and Link',
    long_description='',
    keywords='api, client, plaid',
    author='Plaid Technologies',
    author_email='developers@plaid.com',
    url=url,
    download_url='{}/tarball/v{}'.format(url, version),
    license='MIT',
    packages=find_packages(exclude='tests'),
    package_data={'README': ['README.md']},
    install_requires=['requests>=2.7.0'],
    zip_safe=False,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Environment :: Web Environment",
    ]
)



from plaid.api.accounts import Accounts
from plaid.api.assets import AssetReport
from plaid.api.auth import Auth
from plaid.api.categories import Categories
from plaid.api.credit_details import CreditDetails
from plaid.api.deposit_switch import DepositSwitch
from plaid.api.holdings import Holdings
from plaid.api.identity import Identity
from plaid.api.income import Income
from plaid.api.institutions import Institutions
from plaid.api.investment_transactions import InvestmentTransactions
from plaid.api.item import Item
from plaid.api.liabilities import Liabilities
from plaid.api.link_token import LinkToken
from plaid.api.payment_initiation import PaymentInitiation
from plaid.api.processor import Processor
from plaid.api.sandbox import Sandbox
from plaid.api.transactions import Transactions
from plaid.api.webhooks import Webhooks

__all__ = [
    'Accounts',
    'AssetReport',
    'Auth',
    'Categories',
    'CreditDetails',
    'DepositSwitch',
    'Holdings',
    'Identity',
    'Income',
    'Institutions',
    'InvestmentTransactions',
    'Item',
    'Liabilities',
    'LinkToken',
    'PaymentInitiation',
    'Processor',
    'Sandbox',
    'Transactions',
    'Webhooks',
]



class BaseError(Exception):
    '''
    A base error class.

    :ivar   str     message:            A developer-friendly error message. Not
                                        safe for programmatic use.
    :ivar   str     type:               A broad categorization of the error,
                                        corresponding to the error class.
    :ivar   str     code:               An error code for programmatic use.
    :ivar   str     display_message:    A user-friendly error message. Not safe
                                        for programmatic use. May be None.
    '''
    def __init__(
        self,
        message,
        type,
        code,
        display_message,
    ):
        super(BaseError, self).__init__(message)

        # In Python 3, the Exception class does not expose a `message`
        # attribute so we need to set it explicitly. See
        # https://www.python.org/dev/peps/pep-0352/#retracted-ideas.
        self.message = message

        self.type = type
        self.code = code
        self.display_message = display_message


class PlaidError(BaseError):
    '''
    A Plaid API error.

    :ivar   str     message:            A developer-friendly error message. Not
                                        safe for programmatic use.
    :ivar   str     type:               A broad categorization of the error,
                                        corresponding to the error class.
    :ivar   str     code:               An error code for programmatic use.
    :ivar   str     display_message:    A user-friendly error message. Not safe
                                        for programmatic use. May be None.
    :ivar   str     request_id:         A unique id returned for all server
                                        responses.
    :ivar   list    causes:             A list of reasons explaining why the
                                        error happened.
    '''

    def __init__(
        self,
        message,
        type,
        code,
        display_message,
        request_id="",
        causes=None,
    ):
        super(PlaidError, self).__init__(
            message,
            type,
            code,
            display_message,
        )
        self.request_id = request_id
        self.causes = [
            PlaidCause(
                cause['error_message'],
                cause['error_type'],
                cause['error_code'],
                cause.get('display_message', ''),
                cause['item_id'],
            ) for cause in causes or []
        ]

    @staticmethod
    def from_response(response):
        '''
        Create an error of the right class from an API response.

        :param   response    dict        Response JSON
        '''
        cls = PLAID_ERROR_TYPE_MAP.get(response['error_type'], PlaidError)
        return cls(response['error_message'],
                   response['error_type'],
                   response['error_code'],
                   response['display_message'],
                   response['request_id'],
                   response.get('causes'))


class PlaidCause(BaseError):
    '''
    A cause of a Plaid error.

    :ivar   str     message:            A developer-friendly error message. Not
                                        safe for programmatic use.
    :ivar   str     type:               A broad categorization of the error,
                                        corresponding to the error class.
    :ivar   str     code:               An error code for programmatic use.
    :ivar   str     display_message:    A user-friendly error message. Not safe
                                        for programmatic use. May be None.
    :ivar   str     item_id:            The item ID.
    '''

    def __init__(
        self,
        message,
        type,
        code,
        display_message,
        item_id,
    ):
        super(PlaidCause, self).__init__(
            message,
            type,
            code,
            display_message,
        )
        self.item_id = item_id


class InvalidRequestError(PlaidError):
    '''The request is malformed and cannot be processed.'''

    pass


class InvalidInputError(PlaidError):
    '''The request is correctly formatted, but the values are incorrect.'''

    pass


class RateLimitExceededError(PlaidError):
    '''The request is valid but has exceeded established rate limits.'''

    pass


class APIError(PlaidError):
    '''Planned maintenance or an API internal server error.'''

    pass


class ItemError(PlaidError):
    '''There is invalid information about an item or it is not supported.'''

    pass


class InstitutionError(PlaidError):
    '''There are errors for the requested financial institution.'''

    pass


class AuthError(PlaidError):
    '''There are errors with verifying or pulling auth numbers data.'''

    pass


class AssetReportError(PlaidError):
    '''There are errors with creating the asset report.'''

    pass


PLAID_ERROR_TYPE_MAP = {
    'INSTITUTION_ERROR': InstitutionError,
    'INVALID_REQUEST': InvalidRequestError,
    'INVALID_INPUT': InvalidInputError,
    'RATE_LIMIT_EXCEEDED': RateLimitExceededError,
    'API_ERROR': APIError,
    'ITEM_ERROR': ItemError,
    'AUTH_ERROR': AuthError,
    'ASSET_REPORT_ERROR': AssetReportError,
}



from plaid.api.api import API


class InvestmentTransactions(API):
    '''InvestmentTransactions endpoints.'''

    def get(self,
            access_token,
            start_date,
            end_date,
            _options=None,
            account_ids=None,
            count=None,
            offset=None,
            ):
        '''
        Return accounts and investment transactions for an item.
        (`HTTP docs <https://plaid.com/docs/api/#investment-transactions>`__)

        The transactions in the response are paginated -- compare the number of
        transactions received so far against response['total_transactions'] to
        determine whether to fetch another page.

        :param  str     access_token:
        :param  str     start_date:     The earliest date for transactions.
        :param  str     end_date:       The latest date for transactions.
        :param  [str]   account_ids:    A list of account_ids to retrieve for
                                        the item. Optional.
        :param  int     count:          The number of transactions to fetch.
                                        Optional.
        :param  int     offset:         The number of transactions to skip from
                                        the beginning of the fetch. Optional.

        All date should be formatted as ``YYYY-MM-DD``.
        '''
        options = _options or {}

        if account_ids is not None:
            options['account_ids'] = account_ids
        if count is not None:
            options['count'] = count
        if offset is not None:
            options['offset'] = offset

        return self.client.post('/investments/transactions/get', {
            'access_token': access_token,
            'start_date': start_date,
            'end_date': end_date,
            'options': options,
        })



from plaid.api.api import API


class PaymentInitiation(API):
    '''Payment Initiation endpoints.'''

    def create_recipient(self, name, iban, address):
        '''
        Creates a payment recipient.

        :param  str     name:
        :param  str     iban:
        :param  dict    address: A dictionary containing street, city,
                                 postal_code, and country fields.
        '''

        return self.client.post('/payment_initiation/recipient/create', {
            'name': name,
            'iban': iban,
            'address': address,
        })

    def get_recipient(self, recipient_id):
        '''
        Retrieves a payment recipient.

        :param  str     recipient_id:
        '''

        return self.client.post('/payment_initiation/recipient/get', {
            'recipient_id': recipient_id,
        })

    def list_recipients(self):
        '''
        Lists all the payment recipients.
        '''

        return self.client.post('/payment_initiation/recipient/list', {})

    def create_payment(self, recipient_id, reference, amount):
        '''
        Creates a payment.

        :param  str     recipient_id:
        :param  str     reference:
        :param  dict    amount:       A dictionary containing currency and
                                      value fields.
        '''

        return self.client.post('/payment_initiation/payment/create', {
            'recipient_id': recipient_id,
            'reference': reference,
            'amount': amount,
        })

    def create_payment_token(self, payment_id):
        '''
        Creates a payment.

        :param  str     payment_id:
        '''

        return self.client.post('/payment_initiation/payment/token/create', {
            'payment_id': payment_id,
        })

    def get_payment(self, payment_id):
        '''
        Retrieves a payment.

        :param  str     payment_id:
        '''

        return self.client.post('/payment_initiation/payment/get', {
            'payment_id': payment_id,
        })

    def list_payments(self, options):
        '''
        Lists all the payments.

        :param dict    options:       A dictionary containing count, and
                                      cursor fields.
        '''

        return self.client.post('/payment_initiation/payment/list', options)



from tests.integration.util import (
    create_client,
    SANDBOX_INSTITUTION,
)

access_token = None


def setup_module(module):
    client = create_client()
    pt_response = client.Sandbox.public_token.create(
        SANDBOX_INSTITUTION, ['transactions'])
    exchange_response = client.Item.public_token.exchange(
        pt_response['public_token'])
    global access_token
    access_token = exchange_response['access_token']


def teardown_module(module):
    client = create_client()
    client.Item.remove(access_token)


def test_get():
    client = create_client()

    # get all accounts
    response = client.Accounts.get(access_token)
    assert response['accounts'] is not None

    # get selected accounts
    account_id = response['accounts'][0]['account_id']
    response = client.Accounts.get(access_token, account_ids=[account_id])
    assert len(response['accounts']) == 1


def test_balances_get():
    client = create_client()

    # get all accounts
    response = client.Accounts.balance.get(access_token)
    assert response['accounts'] is not None

    # get selected accounts
    account_id = response['accounts'][0]['account_id']
    response = client.Accounts.balance.get(
        access_token, account_ids=[account_id])
    assert len(response['accounts']) == 1
